;;==========================================================================
;;   SD-
;; !!     2.x !! {   }
;;  r16, r17, r0, r4..r7
;;    C = 1 -  
;;     ffff ffff  (4 ...    2...)
;;  r16  C = 0    (D7/D6/D5/D4):
;; 1xxx -    (   R1,    R1)
;; 0100 -         ( D3..D0    Data Error Token)
;; 0010 -     ( 0x20)
;; 011x -     (D4..D0    Data Response Token)
;; 0000 -   R2 (r0 -   ) [ ]      
;;
;; !!       SD  !!
;; {.. ,             }
;;

.dseg

sd_Type:	.byte   1	; //  : 0 - 1.x;  1 - 2.x, Low;  2 - 2.x, High

sd_OCR:		.byte	4	;  OCR:
; D31 - Card power up status bit (  = 1), D30 - Card Capacity Status (CCS),
; D29..D24 - Reserved,
; D23..D15 -    (  -   0.1   3.5  2.7)

sd_TMP:		.byte	2	;      

.cseg
;
;  SD-
;
sd_Init:
 ldi	r16, spi_CSCLK_release	; !CS = 1
 rcall	spi_CSCLK_Select
 
 ldi	r17, 10			;     MMC,   ,
sd_I_2:				;       
 ldi	r16, 0xFF
 rcall	spi_IO
 dec	r17
 brne	sd_I_2

 ldi	r16, spi_CSCLK_SD	; !CS = 0
 rcall	spi_CSCLK_Select

 ldi	r16, 0xFF		;    
 rcall	spi_IO

;   SPI (RESET): CMD0
; CMD0 [0 0 0 0] 0x95
 clr	r4
 clr	r5
 clr	r6
 clr	r7

 ldi	r16, 0x40		
 ldi	r17, 0x95
 rcall	sd_CmdSend
 andi	r16, 0x01		;   in_idle_state = 1
 breq	sd_Fault2

;     ,       
; (0001b - 2.7..3.6 ),     :      2.x: CMD8
 rcall	sd_Cmd8Send
 andi	r16, 0x04
 brne	sd_I_v1x		;     -     1.x

; ,    v2.x, .
; ģ    idle mode   ,    High Cap: ACMD41

 clr	r4
 clr	r5
 clr	r6
sd_I_1:
 clr	r7

 ldi	r16, 0x77		;  APP_CMD -      
 ldi	r17, 0x65
 rcall	sd_CmdSend		; (CMD55)

 ldi	r16, 0x40		;  HCS = 1
 mov	r7, r16

 ldi	r16, 0x69		;        High Cap
 ldi	r17, 0x77		;         CRC,    -  ?
 rcall	sd_CmdSend		; HC-     "Invalid command"   CRC (  "Invalid CRC" !)

 andi	r16, 0x01		;   in_idle_state = 0
 brne	sd_I_1			;    ,      idle state

;     CCS (Card Capacity Status)

 rcall	sd_Cmd58Send
 ldi	r16, 1
 sts	sd_Type, r16
 lds	r16, sd_OCR+3
 andi	r16, 0x40		; CCS
 breq	sd_I_3

 ldi	r16, 2
 sts	sd_Type, r16
 rjmp	sd_Ok0

;     1.x
;   -    idle state   ACMD41

sd_I_v1x:
 clr	r4
 clr	r5
 clr	r6
 clr	r7			;  HCS = 0
 sts	sd_Type, r7

sd_I_4:
 ldi	r16, 0x77		;  APP_CMD -      
 ldi	r17, 0x65
 rcall	sd_CmdSend		; (CMD55)

 ldi	r16, 0x69		;        High Cap
 ldi	r17, 0xE5
 rcall	sd_CmdSend		;         CRC,    -  ?  ...

 andi	r16, 0x01		;   in_idle_state = 0
 brne	sd_I_4			;    ,      idle state

;   : CMD16 (     Standart Capacity Card)
sd_I_3:
 ldi	r16, 2
 clr	r4
 mov	r5, r16
 clr	r6
 clr	r7
 ldi	r16, 0x50		;     512 
 ldi	r17, 0x15
 rcall	sd_CmdSend

; FIN
sd_Ok0:
 sec
 ret

sd_Fault2:
 clc
 ret

;
;    r4..r7  Z+
;
sd_ReadSect:
 ldi	r16, spi_CSCLK_SD	; !CS = 0
 rcall	spi_CSCLK_Select

 lds	r16, sd_Type
 sbrs	r16, 1
 rcall	sd_Mul512

 ldi	r16, 0x51
 rcall	sd_CmdSendWithoutCRC

sd_RS_2:
 ldi	r16, 0xFF
 rcall	spi_IO

 cpi	r16, 0xFE		;     
 breq	sd_RS_1

 mov	r17, r16
 andi	r17, 0xF0
 brne	sd_RS_2

 ori	r16, 0x40		;    = 0 -    
sd_Fault1:
 rjmp	sd_Fault0

sd_RS_1:
 clr	r17			;    

sd_RS_3:
 ldi	r16, 0xFF
 rcall	spi_IO
 st	Z+, r16

 ldi	r16, 0xFF
 rcall	spi_IO
 st	Z+, r16

 dec	r17
 brne	sd_RS_3

 ldi	r16, 0xFF		;    
 rcall	spi_IO
 ldi	r16, 0xFF
 rcall	spi_IO

sd_Ok1:
 sec
 ret

;
;    r4..r7  Z+
;
sd_WriteSect:
 ldi	r16, spi_CSCLK_SD	; !CS = 0
 rcall	spi_CSCLK_Select

 lds	r16, sd_Type
 sbrs	r16, 1
 rcall	sd_Mul512

 ldi	r16, 0x58
 rcall	sd_CmdSendWithoutCRC

 ldi	r16, 0xFE
 rcall	spi_IO

 clr	r17			;    

sd_WS_3:
 ld	r16, Z+
 rcall	spi_IO

 ld	r16, Z+
 rcall	spi_IO

 dec	r17
 brne	sd_WS_3

 rcall	spi_IO			;    
 rcall	spi_IO

 ldi	r16, 0xFF
 rcall	spi_IO			;   
 andi	r16, 0x1F
 cpi	r16, 0x05
 breq	sd_WS_4			;   

 ori	r16, 0x60		;  ɣ 
 rjmp	sd_Fault0

sd_WS_4:			;   , ģ  
 ldi	r16, 0xFF
 rcall	spi_IO
 cpi	r16, 0xFF
 brne	sd_WS_4

 ldi	r16, 0x4D		;  
 rcall	sd_CmdSendWithoutCRC

 tst	r0
 breq	sd_Ok1

 clr	r16			;  R2,    r0
 rjmp	sd_Fault0

;
;  r4..r7  512
;     sd_*  C = 0
;
sd_Mul512:
 ldi	r16, 0x20	;   

 tst	r7
 brne	sd_CS_3		;    
 tst	r6
 brmi	sd_CS_3		; --""--

 mov	r7, r6
 mov	r6, r5
 mov	r5, r4
 clr	r4

 lsl	r5
 rol	r6
 rol	r7
 ret

;
;      R1 (  r16)  R2 (  r0)
; ,          ,       !
;      R1   ,   sd_*,  C = 0
; "In idle"        r16.
;
;        (CMD0)    "Invalid CRC".
;          .
;
sd_CmdSendWithoutCRC:
 ldi	r17, 0xFF

sd_CmdSend:
 sts	sd_TMP+0, r16
 sts	sd_TMP+1, r17

sd_CS_4:
 ldi	r16, 0xFF		;  
 rcall	spi_IO

 lds	r16, sd_TMP+0
 rcall	spi_IO			;     

 mov	r16, r7			;  
 rcall	spi_IO
 mov	r16, r6
 rcall	spi_IO
 mov	r16, r5
 rcall	spi_IO
 mov	r16, r4
 rcall	spi_IO

 lds	r16, sd_TMP+1		;     
 rcall	spi_IO

 ldi	r16, 0xFF		;   [    
 rcall	spi_IO			;  3F,     R1]

sd_CS_1:
 ldi	r16, 0xFF		;  
 rcall	spi_IO
 tst	r16
 brmi	sd_CS_1

; R1 -    . :
;     D7 == 0, D6 - parameter error, D5 - address error, D4 - erase seq error,
;     D3 - CRC error, D2 - illegal command, D1 - erase reset, D0 - in idle state
; R2 -    .       R1, :
;     D7 - out of range | csd overwrite, D6 - {invalid} erase param, D5 - WP violation,
;     D4 - card ECC failed, D3 - cardcontroller error, D2 - {} error,
;     D1 - WP erase skip | (ul)lock cmd failed, D0 - card is locked

 push	r16			;    R2 -  r0    
 ldi	r16, 0xFF
 rcall	spi_IO
 mov	r0, r16
 pop	r16

 sbrc	r16, 3			;   CRC Error  
 rjmp	sd_CS_4

 mov	r17, r16
 andi	r17, 0x7E
 breq	sd_CS_2

 ori	r16, 0x80		;  

sd_CS_3:
 pop	r17			;          media.SD
 pop	r17

sd_Fault0:
 clc

sd_CS_2:
 ret

;
;   CMD8      R7
;    R1 ( R7)   R16.   .
;
sd_Cmd8Send:
 ldi	r16, 0xFF		;  
 rcall	spi_IO

 ldi	r16, 0x48
 rcall	spi_IO			;     

 clr	r16			;  
 rcall	spi_IO
 clr	r16
 rcall	spi_IO
 ldi	r16, 0x01		; Voltage
 rcall	spi_IO
 ldi	r16, 0xAA		; Check Pattern
 rcall	spi_IO

 ldi	r16, 0x87		;     
 rcall	spi_IO

; ldi	r16, 0xFF		;   [    
; rcall	spi_IO			;  3F,     R1]

sd_C8_1:
 ldi	r16, 0xFF		;  :  :   R1,  ݣ 
 rcall	spi_IO
 tst	r16
 brmi	sd_C8_1

 push	r16
 ldi	r16, 0xFF
 rcall	spi_IO
 ldi	r16, 0xFF
 rcall	spi_IO
 ldi	r16, 0xFF
 rcall	spi_IO
 ldi	r16, 0xFF
 rcall	spi_IO
 pop	r16
 ret

;
;   CMD58      R3
;              C = 0.
;  -  OCR (Operation Conditions Register) -   sd_OCR
;
sd_Cmd58Send:
 ldi	r16, 0xFF		;  
 rcall	spi_IO

 ldi	r16, 0x7A
 rcall	spi_IO			;     

 clr	r16			;   - stuff bits
 rcall	spi_IO
 clr	r16
 rcall	spi_IO
 clr	r16
 rcall	spi_IO
 clr	r16
 rcall	spi_IO

 ldi	r16, 0xFD		;     
 rcall	spi_IO

; ldi	r16, 0xFF		;   [    
; rcall	spi_IO			;  3F,     R1]

sd_C58_1:
 ldi	r16, 0xFF		;  :  :   R1,  ݣ 
 rcall	spi_IO
 tst	r16
 brmi	sd_C58_1
 brne	sd_CS_3

 ldi	r16, 0xFF
 rcall	spi_IO
 sts	sd_OCR+3, r16
 ldi	r16, 0xFF
 rcall	spi_IO
 sts	sd_OCR+2, r16
 ldi	r16, 0xFF
 rcall	spi_IO
 sts	sd_OCR+1, r16
 ldi	r16, 0xFF
 rcall	spi_IO
 sts	sd_OCR+0, r16
 ret
